#include<opencv2/opencv.hpp>
#include<opencv2/highgui.hpp>
#include "opencv2/imgproc.hpp"
#include<iostream>
#include <chrono>
#include <thread>

#define _WINSOCK_DEPRECATED_NO_WARNINGS

#ifdef _WIN32
#include <Winsock2.h> // before Windows.h, else Winsock 1 conflict
#include <Ws2tcpip.h> // needed for ip_mreq definition for multicast
#include <Windows.h>
#else
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <time.h>
#endif

#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#define _USE_MATH_DEFINES
#include <math.h>


class expAvg {
private:
	float w;
	float avg;
	float lambda;
public:
	expAvg(float lambda) : lambda(lambda) {
		w = 1.0f;
		avg = 0.0f;
	};
	void insertVal(float val) {
		w = lambda*w + 1;
		avg = (1 - 1 / w)*avg + (1 / w)*val;
	};
	float getVal() {
		return avg;
	};
};



//using namespace std;  // conflict with WINSOCK API
using namespace cv;

/* Sliders and other control */
const int alpha_slider_max = 100;
const int beta_slider_max = 100;
int alpha_slider = 0;
int beta_slider = 1000000;
int crop_slider = 1;
double alpha = 10;
double beta = 0;
double gamma = 0;
int video_slider = 0;
int calib_slider = 0;

/* dimensions of output of remap*/
#define REMAP_ROWS 400
#define REMAP_COLS (2*REMAP_ROWS)

/*
 * Prepare maps
 */
Mat mapx = Mat(REMAP_ROWS, REMAP_COLS, CV_32FC1, Scalar(0));
Mat mapy = Mat(REMAP_ROWS, REMAP_COLS, CV_32FC1, Scalar(0));

/*
 * Converted maps for better performance
 */
Mat map1 = Mat(REMAP_ROWS, REMAP_COLS, CV_16SC2);
Mat map2 = Mat(REMAP_ROWS, REMAP_COLS, CV_16UC1);

/*Intermediate matrices to be composed to output frame*/
Mat imrh, imrv;
Mat colorImgh, colorImgv;
Mat frframe;
Mat frame;
Mat rframe;
Mat black = Mat(Size(REMAP_COLS, REMAP_ROWS), CV_8UC3, Scalar(0));

Mat hc, hc2;
Mat colorImgvr;

expAvg avgFrameRate(0.99f);


/* variables for calibration process*/
std::ofstream f;
double sumx, sumy, sumxy, sumx2;
unsigned n;

/* alpha slider callback*/
static void on_trackbar1(int, void*)
{
	alpha = (double)alpha_slider;
}

/*beta slide callback*/
static void on_trackbar2(int, void*)
{
	beta = (double)beta_slider;
}



int width = 0;
int height = 0;
float fi_0 = 0;
//bool first_crop = true;

/* compute map for remap fuction, conversion from rectangle to polar, crop slider callback*/
static void on_trackbar3(int, void*)
{
	Point my_o = Point(REMAP_COLS/ 2, 0);
	for (int row = 0; row < REMAP_ROWS; row++) {
		for (int col = 0; col < REMAP_COLS; col++) {
			/* generate image with center top middle */
			/* actual point */
			Point pt = Point(col, row);
			/* relative to my origin */
			Point rpt = pt - my_o;

			/* compute angle to vertical axis, positive anticlockwise, zero down*/
			float d = sqrtf((float)rpt.x*(float)rpt.x + (float)rpt.y*(float)rpt.y);
			float fi = asinf((float)rpt.x / d);

			/* fi to degrees*/
			fi = fi / ((float)M_PI / 180);

			/* fi to index to source matrix*/
			float fi_idx = (fi_0 - fi) / (2*fi_0/(float)(height-1));
			float  d_idx = width * d / REMAP_ROWS;

			mapx.at<float>(row, col) = d_idx < crop_slider ? -1 : d_idx > width ? -1 : d_idx;
			mapy.at<float>(row, col) = fi_idx < 0 ? -1 : fi_idx > (height-1) ? -1 : fi_idx;

		}
	}
	/* convert maps for better performance*/
	convertMaps(mapx, mapy, map1, map2, CV_16SC2);
}

VideoWriter *video = NULL;

/* callback for video recording switch (0-1 slider)*/
static void on_video_slider(int, void*) {
	switch (video_slider) {
	case 1: /* start writing video */
			// Define the codec and create VideoWriter object.The output is stored in 'out.avi' file. 
			video = new VideoWriter("out.avi", CV_FOURCC('M', 'J', 'P', 'G'), avgFrameRate.getVal(), hc2.size());
			break;
	case 0: /* stop video writing */
			video->release();
			break;
	}
}

/* callback for calibration switch (0-1 slider)*/
static void on_calib_slider(int, void*) {
	switch (calib_slider) {
	case 1:  // start calibration process
		sumx = sumy = sumxy = sumx2 = 0;
		n = 0;
		f.open("calib.csv");
		break;
	case 0:
		float A = (float)(((sumx2*sumy - sumx*sumxy)*1.0f / (n*sumx2 - sumx*sumx)*1.0));
		float b = (float)(((n*sumxy - sumx*sumy)*1.0 / (n*sumx2 - sumx*sumx)*1.0));
		float a = expf(A);
		f << std::endl;
		f << std::endl;
		f << "The normalization curve is Y = " << a << ".e^" << b << std::endl;
		f.close();
		break;
	}
}

int flip_val = 10;
int flip_slider;
bool camera_off = false;
/* callback for flipping the camera image*/
static void on_flip1(int, void*) {
	camera_off = false;
	switch (flip_slider) {
	case 0: 
		flip_val = 10;
		break;
	case 1: 
		flip_val = 0;
		break;
	case 2:
		flip_val = 1;
		break;
	case 3:
		flip_val = -1;
		break;
	case 4:
		camera_off = true;
		break;
	default:
		flip_val = 10;
		break;
	}
}

/*rotate point 2d around origin*/
cv::Point2f rotate2d(const cv::Point2f& inPoint, const float& angRad)
{
	cv::Point2f outPoint;
	//CW rotation
	outPoint.x = std::cosf(angRad)*inPoint.x - std::sinf(angRad)*inPoint.y;
	outPoint.y = std::sinf(angRad)*inPoint.x + std::cosf(angRad)*inPoint.y;
	return outPoint;
}

/*rotate point around centerpoint */
cv::Point2f rotatePoint(const cv::Point2f& inPoint, const cv::Point2f& center, const float& angRad)
{
	return rotate2d(inPoint - center, angRad) + center;
}


bool capture_thread = true;
unsigned char cap_data[2][3][65535];
volatile unsigned char cap_data_lock = 0;
volatile unsigned char proc_data_lock = 0;
int fd;
struct sockaddr_in addr;
int addrlen = sizeof(addr);

void eth_capture_thread(void) {
	std::clog << "Starting packet capture from ethernet" << std::endl;

	while (capture_thread) {
		/* do packet capture */
		int nbytes = recvfrom(
			fd,
			(char*)cap_data[0][cap_data_lock],
			65535,
			0,
			(struct sockaddr *) &addr,
			&addrlen
		);
		if (nbytes == SOCKET_ERROR)
		{
			int err = WSAGetLastError();

			if (WSAGetLastError() == WSAETIMEDOUT) {
				perror("recvfrom: timed out");
			}
			else {
				perror("recvfrom: other error: code "); printf("%d\r\n", err);
			}
			std::this_thread::sleep_for(std::chrono::seconds(1));
		}

		if (cap_data[0][cap_data_lock][1] != 0) {
			// Resync
			std::clog << "Skipping one packet to get sync" << std::endl;
			continue;
		}

		/* do packet capture */
		nbytes = recvfrom(
			fd,
			(char*)cap_data[1][cap_data_lock],
			65535,
			0,
			(struct sockaddr *) &addr,
			&addrlen
		);
		if (nbytes == SOCKET_ERROR)
		{
			int err = WSAGetLastError();

			if (WSAGetLastError() == WSAETIMEDOUT) {
				perror("recvfrom: timed out");
			}
			else {
				perror("recvfrom: other error: code "); printf("%d\r\n", err);
			}
			std::this_thread::sleep_for(std::chrono::seconds(1));
		}

		if ((cap_data[0][cap_data_lock][0] == cap_data[1][cap_data_lock][0]) && (cap_data[1][cap_data_lock][1] == 1)) {
			// now i have collected horizontal and vertical measurement
			int new_cap_data_lock = (cap_data_lock == 2) ? 0 : cap_data_lock + 1;
			if (new_cap_data_lock != proc_data_lock) {
				cap_data_lock = new_cap_data_lock;
			}
		}
		else {
			std::clog << "Wrong data pair, skipping ... " << std::endl;
		}

	}
}

int main()
{
	char group[50] = "239.0.0.1"; // multicast group to receive from
	int port = 6001;              // port, 0 if error, which is an invalid port
	unsigned char *msgbufh, *msgbufv; // packet buffers horizontal and vertical

#ifdef _WIN32
	//
	// Initialize Windows Socket API with given VERSION.
	//
	WSADATA wsaData;
	std::clog << "WSA Startup" << std::endl;
	if (WSAStartup(0x0101, &wsaData)) {
		perror("WSAStartup");
		return 1;
	}
#endif

	// create what looks like an ordinary UDP socket
	//
	std::clog << "Opening socket" << std::endl;
	fd = (int)socket(AF_INET, SOCK_DGRAM, 0);
	if (fd < 0) {
		perror("socket");
		return 1;
	}

	// allow multiple sockets to use the same PORT number
	//
	u_int yes = 1;
	std::clog << "Socket set to reuse port number" << std::endl;
	if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (char*)&yes, sizeof(yes)) < 0 ) {
		perror("Reusing ADDR failed");
		return 1;
	}

	// set up destination address
	//
	
	memset(&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = htonl(INADDR_ANY); // differs from sender
	addr.sin_port = htons(port);

	// bind to receive address
	//
	std::clog << "Bind to socket" << std::endl;
	if (bind(fd, (struct sockaddr*) &addr, sizeof(addr)) < 0) {
		perror("bind");
		return 1;
	}
	
	// use setsockopt() to request that the kernel join a multicast group
	//
	struct ip_mreq mreq;
	mreq.imr_multiaddr.s_addr = inet_addr(group);
	mreq.imr_interface.s_addr = htonl(INADDR_ANY);
	std::clog << "Socket set to join multicast group" << std::endl;
	if (setsockopt(fd, IPPROTO_IP, IP_ADD_MEMBERSHIP, (char*)&mreq, sizeof(mreq)) < 0 ) {
		perror("setsockopt");
		return 1;
	}

	DWORD timeout = 3*1000;  /* 3 seconds timeout */
	std::clog << "Socket set to timeout if no data available" << std::endl;
	setsockopt(fd, SOL_SOCKET, SO_RCVTIMEO, (char*)&timeout, sizeof(timeout));

	Mat img;
	/* create window and sliders, uncomment all needed sliders*/
	std::clog << "Creating output window and sliders" << std::endl;
	namedWindow("image", WINDOW_NORMAL/*WINDOW_NORMAL | WINDOW_KEEPRATIO*/);
//	createTrackbar("Crop", "image", &crop_slider, 500, on_trackbar3); 
	crop_slider = 220;
	createTrackbar("Video","image", &video_slider, 1, on_video_slider);
//	createTrackbar("Calib", "image", &calib_slider, 1, on_calib_slider);
//	createTrackbar("Flip", "image", &flip_slider, 4, on_flip1);

	std::clog << "Capture device init" << std::endl;
	/* capture form camera if available*/
	VideoCapture cap(0); // open the default camera

	


	std::thread t1(eth_capture_thread);

	while (1) {
		
		// Start time
		int64 start = cv::getTickCount();

		while (proc_data_lock == cap_data_lock) {};

		msgbufh = cap_data[0][proc_data_lock];
		msgbufv = cap_data[1][proc_data_lock];

		/* parse packet headers */
		//std::clog << "Parsing packet header" << std::endl;
		int packet_group_id = msgbufh[0];
		int packet_data_id = msgbufh[1];
		int rows = (msgbufh[3] << 8) + msgbufh[2];
		int cols = (msgbufh[5] << 8) + msgbufh[4];
		float fi0 = *((float*)&msgbufh[6]);
		unsigned char cps = msgbufh[10];
		unsigned hmax_val = msgbufh[14]*(1 << 24) + msgbufh[13]*(1 << 16) + msgbufh[12]*(1 << 8) + msgbufh[11];
		unsigned hmax_row = msgbufh[16]*(1 << 8) + msgbufh[15];
		unsigned hmax_col = msgbufh[18]*(1 << 8) + msgbufh[17];
		unsigned vmax_val = msgbufv[14] * (1 << 24) + msgbufv[13] * (1 << 16) + msgbufv[12] * (1 << 8) + msgbufv[11];
		unsigned vmax_row = msgbufv[16] * (1 << 8) + msgbufv[15];
		unsigned vmax_col = msgbufv[18] * (1 << 8) + msgbufv[17];

		/*
		* compute coordinates of maximum in horizontal and vertical image
		*/
		Point hrpt,vrpt;
		float rfi;
		//std::clog << "Projection of point of global max to polar plot" << std::endl;
		if (width != 0) {
			Point hpt = Point(REMAP_ROWS*hmax_col / width, 0);
			rfi = (fi0 + (hmax_row * ((-2 * fi0) / (rows - 1))))*((float)M_PI / 180);
			hrpt = rotate2d(hpt, rfi);
			hrpt.y += REMAP_COLS / 2;
			Point vpt = Point(REMAP_ROWS*vmax_col / width, 0);
			rfi = (fi0 + (vmax_row * ((-2 * fi0) / (rows - 1))))*((float)M_PI / 180);
			vrpt = rotate2d(vpt, rfi);
			vrpt.y += REMAP_COLS / 2;
		}


    	/*
		* update image parameters and recompute remap matrices after dimensions has been changed
		*/
		if ((width != cols) || (height != rows) || (fi_0 != fi0)) {
			width = cols;
			height = rows;
			fi_0 = fi0;
			on_trackbar3(33, 0);
			resizeWindow("image", 3*REMAP_COLS/2, 2*REMAP_ROWS);// # resize window
		}

		// Setup mats with received data */
		Mat imgh = Mat(height, width, CV_8UC1, (void*)(&msgbufh[19]));
		Mat imgv = Mat(height, width, CV_8UC1, (void*)(&msgbufv[19]));

		/* remap data to polar using converted remap matrices */
		remap(imgh, imrh, map1, map2, INTER_LINEAR, BORDER_CONSTANT, Scalar(0, 0, 0));
		remap(imgv, imrv, map1, map2, INTER_LINEAR, BORDER_CONSTANT, Scalar(0, 0, 0));

		/* add colors */
		applyColorMap(imrh, colorImgh, COLORMAP_JET);
		applyColorMap(imrv, colorImgv, COLORMAP_JET);

		/* add USB camera frame if available, black otherwise*/
		if (cap.isOpened() && (!camera_off)) {

			cap >> frame;
			resize(frame, rframe, Size(REMAP_COLS, REMAP_ROWS));

			if (flip_val != 10) {
				flip(rframe, frframe, flip_val);
			}
			else {
				frframe = rframe;
			}
		}
		else { // no USB camera present, insert black rectangle
			frframe = black;
		}

		/*
		* mark point of maxima by white rectangle in horizontal and vertical image
		*/
		Rect hrect(hrpt.y - 5, hrpt.x - 5, 10, 10);
		Rect vrect(vrpt.y - 5, vrpt.x - 5, 10, 10);

		rectangle(colorImgh, hrect, Scalar(255, 255, 255), 2, 8, 0);
		rectangle(colorImgv, vrect, Scalar(255, 255, 255), 2, 8, 0);

		/* put outputs together to one image*/
		vconcat(frframe, colorImgh, hc);
		rotate(colorImgv, colorImgvr, cv::ROTATE_90_CLOCKWISE);
		hconcat(colorImgvr, hc, hc2);

		imshow("image", hc2);

		/* exit on ESC and close video and calib file if still writing them*/
		if ((waitKey(16) & 0xff) == 27) {
			if (video_slider == 1) {
				video_slider = 0;
				on_video_slider(0,NULL);
			}
			if (calib_slider == 1) {
				calib_slider = 0;
				on_calib_slider(0, NULL);
			}

			capture_thread = false;
			t1.join();
			break;
		}

		// video write frame 
		if (video_slider == 1) {
			video->write(hc2);
		}

		// fps computation
		double fps = cv::getTickFrequency() / (cv::getTickCount() - start);
		avgFrameRate.insertVal((float)fps);
		printf("Data set %d at FPS: %3.1f\r\n", msgbufh[0], avgFrameRate.getVal());

		/* capture calibration data*/
		if (calib_slider == 1) {
			f << hmax_col << " ; " << hmax_val << std::endl;
			sumx  += hmax_col;
			sumx2 += hmax_col*hmax_col;
			sumy  += log(hmax_val);
			sumxy += hmax_col*log(hmax_val);
			n++;
		}

		/* move to new proc data*/
		proc_data_lock = (proc_data_lock == 2) ? 0 : proc_data_lock + 1;
	}


#ifdef _WIN32
	//
	// Program never actually gets here due to infinite loop that has to be
	// canceled, but since people on the internet wind up using examples
	// they find at random in their own code it's good to show what shutting
	// down cleanly would look like.
	//
	WSACleanup();
#endif
}